// Code generated by "stringer -type=SymKind"; DO NOT EDIT.

package sym

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[Sxxx-0]
	_ = x[STEXT-1]
	_ = x[STEXTFIPSSTART-2]
	_ = x[STEXTFIPS-3]
	_ = x[STEXTFIPSEND-4]
	_ = x[STEXTEND-5]
	_ = x[SELFRXSECT-6]
	_ = x[SMACHOPLT-7]
	_ = x[STYPE-8]
	_ = x[SSTRING-9]
	_ = x[SGOSTRING-10]
	_ = x[SGOFUNC-11]
	_ = x[SGCBITS-12]
	_ = x[SRODATA-13]
	_ = x[SRODATAFIPSSTART-14]
	_ = x[SRODATAFIPS-15]
	_ = x[SRODATAFIPSEND-16]
	_ = x[SRODATAEND-17]
	_ = x[SFUNCTAB-18]
	_ = x[SELFROSECT-19]
	_ = x[STYPERELRO-20]
	_ = x[SSTRINGRELRO-21]
	_ = x[SGOSTRINGRELRO-22]
	_ = x[SGOFUNCRELRO-23]
	_ = x[SGCBITSRELRO-24]
	_ = x[SRODATARELRO-25]
	_ = x[SFUNCTABRELRO-26]
	_ = x[SELFRELROSECT-27]
	_ = x[STYPELINK-28]
	_ = x[SITABLINK-29]
	_ = x[SSYMTAB-30]
	_ = x[SPCLNTAB-31]
	_ = x[SFirstWritable-32]
	_ = x[SBUILDINFO-33]
	_ = x[SFIPSINFO-34]
	_ = x[SELFSECT-35]
	_ = x[SMACHO-36]
	_ = x[SMACHOGOT-37]
	_ = x[SWINDOWS-38]
	_ = x[SELFGOT-39]
	_ = x[SNOPTRDATA-40]
	_ = x[SNOPTRDATAFIPSSTART-41]
	_ = x[SNOPTRDATAFIPS-42]
	_ = x[SNOPTRDATAFIPSEND-43]
	_ = x[SNOPTRDATAEND-44]
	_ = x[SINITARR-45]
	_ = x[SDATA-46]
	_ = x[SDATAFIPSSTART-47]
	_ = x[SDATAFIPS-48]
	_ = x[SDATAFIPSEND-49]
	_ = x[SDATAEND-50]
	_ = x[SXCOFFTOC-51]
	_ = x[SBSS-52]
	_ = x[SNOPTRBSS-53]
	_ = x[SLIBFUZZER_8BIT_COUNTER-54]
	_ = x[SCOVERAGE_COUNTER-55]
	_ = x[SCOVERAGE_AUXVAR-56]
	_ = x[STLSBSS-57]
	_ = x[SXREF-58]
	_ = x[SMACHOSYMSTR-59]
	_ = x[SMACHOSYMTAB-60]
	_ = x[SMACHOINDIRECTPLT-61]
	_ = x[SMACHOINDIRECTGOT-62]
	_ = x[SFILEPATH-63]
	_ = x[SDYNIMPORT-64]
	_ = x[SHOSTOBJ-65]
	_ = x[SUNDEFEXT-66]
	_ = x[SDWARFSECT-67]
	_ = x[SDWARFCUINFO-68]
	_ = x[SDWARFCONST-69]
	_ = x[SDWARFFCN-70]
	_ = x[SDWARFABSFCN-71]
	_ = x[SDWARFTYPE-72]
	_ = x[SDWARFVAR-73]
	_ = x[SDWARFRANGE-74]
	_ = x[SDWARFLOC-75]
	_ = x[SDWARFLINES-76]
	_ = x[SSEHUNWINDINFO-77]
	_ = x[SSEHSECT-78]
}

const _SymKind_name = "SxxxSTEXTSTEXTFIPSSTARTSTEXTFIPSSTEXTFIPSENDSTEXTENDSELFRXSECTSMACHOPLTSTYPESSTRINGSGOSTRINGSGOFUNCSGCBITSSRODATASRODATAFIPSSTARTSRODATAFIPSSRODATAFIPSENDSRODATAENDSFUNCTABSELFROSECTSTYPERELROSSTRINGRELROSGOSTRINGRELROSGOFUNCRELROSGCBITSRELROSRODATARELROSFUNCTABRELROSELFRELROSECTSTYPELINKSITABLINKSSYMTABSPCLNTABSFirstWritableSBUILDINFOSFIPSINFOSELFSECTSMACHOSMACHOGOTSWINDOWSSELFGOTSNOPTRDATASNOPTRDATAFIPSSTARTSNOPTRDATAFIPSSNOPTRDATAFIPSENDSNOPTRDATAENDSINITARRSDATASDATAFIPSSTARTSDATAFIPSSDATAFIPSENDSDATAENDSXCOFFTOCSBSSSNOPTRBSSSLIBFUZZER_8BIT_COUNTERSCOVERAGE_COUNTERSCOVERAGE_AUXVARSTLSBSSSXREFSMACHOSYMSTRSMACHOSYMTABSMACHOINDIRECTPLTSMACHOINDIRECTGOTSFILEPATHSDYNIMPORTSHOSTOBJSUNDEFEXTSDWARFSECTSDWARFCUINFOSDWARFCONSTSDWARFFCNSDWARFABSFCNSDWARFTYPESDWARFVARSDWARFRANGESDWARFLOCSDWARFLINESSSEHUNWINDINFOSSEHSECT"

var _SymKind_index = [...]uint16{0, 4, 9, 23, 32, 44, 52, 62, 71, 76, 83, 92, 99, 106, 113, 129, 140, 154, 164, 172, 182, 192, 204, 218, 230, 242, 254, 267, 280, 289, 298, 305, 313, 327, 337, 346, 354, 360, 369, 377, 384, 394, 413, 427, 444, 457, 465, 470, 484, 493, 505, 513, 522, 526, 535, 558, 575, 591, 598, 603, 615, 627, 644, 661, 670, 680, 688, 697, 707, 719, 730, 739, 751, 761, 770, 781, 790, 801, 815, 823}

func (i SymKind) String() string {
	if i >= SymKind(len(_SymKind_index)-1) {
		return "SymKind(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _SymKind_name[_SymKind_index[i]:_SymKind_index[i+1]]
}
